// Copyright 2020 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "generator/internal/client_generator.h"
#include "google/cloud/internal/absl_str_cat_quiet.h"
#include "absl/memory/memory.h"
#include "generator/internal/codegen_utils.h"
#include "generator/internal/descriptor_utils.h"
#include "generator/internal/predicate_utils.h"
#include "generator/internal/printer.h"
#include <google/api/client.pb.h>
#include <google/protobuf/descriptor.h>

namespace google {
namespace cloud {
namespace generator_internal {

ClientGenerator::ClientGenerator(
    google::protobuf::ServiceDescriptor const* service_descriptor,
    VarsDictionary service_vars,
    std::map<std::string, VarsDictionary> service_method_vars,
    google::protobuf::compiler::GeneratorContext* context)
    : ServiceCodeGenerator("client_header_path", "client_cc_path",
                           service_descriptor, std::move(service_vars),
                           std::move(service_method_vars), context) {}

Status ClientGenerator::GenerateHeader() {
  HeaderPrint(CopyrightLicenseFileHeader());
  HeaderPrint(  // clang-format off
    "// Generated by the Codegen C++ plugin.\n"
    "// If you make any local changes, they will be lost.\n"
    "// source: $proto_file_name$\n"
    "#ifndef $header_include_guard$\n"
    "#define $header_include_guard$\n"
    "\n");
  // clang-format on

  // includes
  HeaderLocalIncludes({vars("connection_header_path"), "google/cloud/future.h",
                       "google/cloud/polling_policy.h",
                       "google/cloud/status_or.h", "google/cloud/version.h"});
  HeaderSystemIncludes({"google/longrunning/operations.grpc.pb.h", "memory"});
  HeaderPrint("\n");

  auto result = HeaderOpenNamespaces();
  if (!result.ok()) return result;

  // Client Class
  HeaderPrint(  // clang-format off
    "$class_comment_block$\n"
    "class $client_class_name$ {\n"
    " public:\n"
    "  explicit $client_class_name$(ConnectionOptions const& options = ConnectionOptions());\n"
    "  explicit $client_class_name$(std::shared_ptr<$connection_class_name$> connection);\n"
    "  ~$client_class_name$();\n"
    "\n"
    "  //@{\n"
    "  // @name Copy and move support\n"
    "  $client_class_name$($client_class_name$ const&) = default;\n"
    "  $client_class_name$& operator=($client_class_name$ const&) = default;\n"
    "  $client_class_name$($client_class_name$&&) noexcept = default;\n"
    "  $client_class_name$& operator=($client_class_name$&&) noexcept = default;\n"
    "  //@}\n"
    "\n"
    "  //@{\n"
    "  // @name Equality\n"
    "  friend bool operator==($client_class_name$ const& a, $client_class_name$ const& b) {\n"
    "    return a.connection_ == b.connection_;\n"
    "  }\n"
    "  friend bool operator!=($client_class_name$ const& a, $client_class_name$ const& b) {\n"
    "    return !(a == b);\n"
    "  }\n"
    "  //@}\n"
    "\n");
  // clang-format on

  for (google::protobuf::MethodDescriptor const& method : methods()) {
    auto method_signature_extension =
        method.options().GetRepeatedExtension(google::api::method_signature);
    for (int i = 0; i < method_signature_extension.size(); ++i) {
      std::string method_string =
          absl::StrCat("  $method_name$($method_signature", i, "$);\n\n");
      HeaderPrintMethod(
          method,
          {MethodPattern(
               {
                   {"  $method_signature_comment_block$\n"},
                   {IsResponseTypeEmpty,
                    // clang-format off
                   "  Status\n",
                   "  StatusOr<$response_type$>\n"},
                  {method_string}
                   // clang-format on
               },
               All(IsNonStreaming, Not(IsLongrunningOperation),
                   Not(IsPaginated))),
           MethodPattern(
               {
                   {"  $method_signature_comment_block$\n"},
                   {IsResponseTypeEmpty,
                    // clang-format off
                    "  future<Status>\n",
                    "  future<StatusOr<$longrunning_deduced_response_type$>>\n"},
                   {method_string}
                   // clang-format on
               },
               All(IsNonStreaming, IsLongrunningOperation, Not(IsPaginated))),
           MethodPattern(
               {
                   // clang-format off
                   {"  $method_signature_comment_block$\n"
                    "  $method_name$Range\n"},
                   {method_string},
                   // clang-format on
               },
               All(IsNonStreaming, Not(IsLongrunningOperation), IsPaginated))},
          __FILE__, __LINE__);
    }
  }

  for (auto const& method : methods()) {
    HeaderPrintMethod(
        method,
        {MethodPattern(
             {
                 {"  $request_comment_block$\n"},
                 {IsResponseTypeEmpty,
                  // clang-format off
    "  Status\n",
    "  StatusOr<$response_type$>\n"},
   {"  $method_name$($request_type$ const& request);\n"
        "\n"}
                 // clang-format on
             },
             All(IsNonStreaming, Not(IsLongrunningOperation),
                 Not(IsPaginated))),
         MethodPattern(
             {
                 {"  $request_comment_block$\n"},
                 {IsResponseTypeEmpty,
                  // clang-format off
    "  future<Status>\n",
    "  future<StatusOr<$longrunning_deduced_response_type$>>\n"},
   {"  $method_name$($request_type$ const& request);\n"
        "\n"}
                 // clang-format on
             },
             All(IsNonStreaming, IsLongrunningOperation, Not(IsPaginated))),
         MethodPattern(
             {
                 // clang-format off
                  {"  $request_comment_block$\n"},
   {"  $method_name$Range\n"
    "  $method_name$($request_type$ request);\n\n"},
                 // clang-format on
             },
             All(IsNonStreaming, Not(IsLongrunningOperation), IsPaginated))},
        __FILE__, __LINE__);
  }

  HeaderPrint(  // clang-format off
    " private:\n"
    "  std::shared_ptr<$connection_class_name$> connection_;\n");
  // clang-format on

  // close Client class
  HeaderPrint(  // clang-format off
    "};\n\n");
  // clang-format on

  HeaderCloseNamespaces();
  // close header guard
  HeaderPrint(  // clang-format off
    "#endif  // $header_include_guard$\n");
  // clang-format on
  return {};
}

Status ClientGenerator::GenerateCc() {
  CcPrint(CopyrightLicenseFileHeader());
  CcPrint(  // clang-format off
    "// Generated by the Codegen C++ plugin.\n"
    "// If you make any local changes, they will be lost.\n"
    "// source: $proto_file_name$\n\n");
  // clang-format on

  // includes
  CcLocalIncludes({vars("client_header_path")});
  CcSystemIncludes({"memory"});
  CcPrint("\n");

  auto result = CcOpenNamespaces();
  if (!result.ok()) return result;

  CcPrint(  // clang-format off
    "$client_class_name$::$client_class_name$(ConnectionOptions const& options)\n"
    " : connection_(Make$connection_class_name$(options)) {}\n");
  // clang-format on

  CcPrint(  // clang-format off
    "$client_class_name$::$client_class_name$(std::shared_ptr<$connection_class_name$> connection)"
    " : connection_(std::move(connection)) {}\n");
  // clang-format on

  CcPrint(  // clang-format off
    "$client_class_name$::~$client_class_name$() = default;\n\n");
  // clang-format on

  for (google::protobuf::MethodDescriptor const& method : methods()) {
    auto method_signature_extension =
        method.options().GetRepeatedExtension(google::api::method_signature);
    for (int i = 0; i < method_signature_extension.size(); ++i) {
      std::string method_string = absl::StrCat(
          "$client_class_name$::$method_name$($method_signature", i, "$) {\n");
      std::string method_request_string =
          absl::StrCat("$method_request_setters", i, "$");
      CcPrintMethod(
          method,
          {MethodPattern(
               {
                   {IsResponseTypeEmpty,
                    // clang-format off
                   "Status\n",
                   "StatusOr<$response_type$>\n"},
                  {method_string},
                  {"  $request_type$ request;\n"},
                   {method_request_string},
                  {"  return connection_->$method_name$(request);\n"
                   "}\n\n"}
                   // clang-format on
               },
               All(IsNonStreaming, Not(IsLongrunningOperation),
                   Not(IsPaginated))),
           MethodPattern(
               {
                   {IsResponseTypeEmpty,
                    // clang-format off
                    "future<Status>\n",
                    "future<StatusOr<$longrunning_deduced_response_type$>>\n"},
                  {method_string},
                  {"  $request_type$ request;\n"},
                   {method_request_string},
                  {"  return connection_->$method_name$(request);\n"
                  "}\n\n"}
                   // clang-format on
               },
               All(IsNonStreaming, IsLongrunningOperation, Not(IsPaginated))),
           MethodPattern(
               {
                   // clang-format off
                   {"$method_name$Range\n"},
                  {method_string},
                  {"  $request_type$ request;\n"},
                   {method_request_string},
                  {"  return connection_->$method_name$(request);\n"
                  "}\n\n"}
                   // clang-format on
               },
               All(IsNonStreaming, Not(IsLongrunningOperation), IsPaginated))},
          __FILE__, __LINE__);
    }
  }

  for (auto const& method : methods()) {
    CcPrintMethod(
        method,
        {MethodPattern(
             {
                 {IsResponseTypeEmpty,
                  // clang-format off
    "Status\n",
    "StatusOr<$response_type$>\n"},
   {"$client_class_name$::$method_name$($request_type$ const& request) {\n"
    "  return connection_->$method_name$(request);\n"
    "}\n\n"}
                 // clang-format on
             },
             All(IsNonStreaming, Not(IsLongrunningOperation),
                 Not(IsPaginated))),
         MethodPattern(
             {
                 {IsResponseTypeEmpty,
                  // clang-format off
    "future<Status>\n",
    "future<StatusOr<$longrunning_deduced_response_type$>>\n"},
   {"$client_class_name$::$method_name$($request_type$ const& request) {\n"
    "  return connection_->$method_name$(request);\n"
    "}\n\n"}
                 // clang-format on
             },
             All(IsNonStreaming, IsLongrunningOperation, Not(IsPaginated))),
         MethodPattern(
             {
                 // clang-format off
   {"$method_name$Range\n"
    "$client_class_name$::$method_name$($request_type$ request) {\n"
    "  return connection_->$method_name$(request);\n"
    "}\n\n"}
                 // clang-format on
             },
             All(IsNonStreaming, Not(IsLongrunningOperation), IsPaginated))},
        __FILE__, __LINE__);
  }

  CcCloseNamespaces();
  return {};
}

}  // namespace generator_internal
}  // namespace cloud
}  // namespace google
